# BEGIN COPYRIGHT
# *************************************************************************
# Licensed Materials - Property of IBM 
# 5724-L01, 5655-N53, 5724-I82, 5655-R15
# Copyright IBM Corporation 2007. All rights reserved. 
# US Government Users Restricted Rights - Use, duplication, or disclosure
# restricted by GSA ADP Schedule Contract with IBM Corp.
# *************************************************************************
# END COPYRIGHT

# import related classes
import sys
from java.lang import *
from com.ibm.db2j.jdbc import *
from java.sql import *
import java.io.File as File
from javax.xml.parsers import *

#-----------------------------------------------------------------
# Usage
#
#-----------------------------------------------------------------
def usage():
    print "RelServiceUpgrade.py is used to upgrade Cloudscape5.1 instances that are used "\
          "by WPS60x profiles. This script is invoked during a profile upgrade. Also, you "\
          "may invoke it to upgrade a Cloudscape5.1 instance manually."
    print "Usage:"  
    print "<wsadmin_commond> -lang jython -conntype NONE " \
          "-wsadmin_classpath <WPS60x_HOME>/cloudscape/lib/db2j.jar" \
          "-f RelServiceUpgrade.py " \
            "[<WPS60x_PROFILE_HOME> <backup_directory>] " \
            "| [-manuallyUpgrade <databaseURL>]" \

    print "Where:"
    print "    In windows, <wsadmin_command> is wsadmin.bat"
    print "    In Linux/Unix, <wsadmin_command> is wsadmin.sh"
    print "    <WPS60x_PROFILE_HOME> is the path to the source WPS60x profile home directory"
    print "    <backup_directory> is the path to the backup directory"
    print "    <databaseURL> is the path to the Cloudscape 5.1 instance that needs manually upgrade"
    print "    Please note:"
    print "        Please pass in valid file paths."
    print "            In windows, the file path should be separated by \\\\"
    print ""
    print "Examples:"
    print "1. To invoke this script during profile pre-upgrade:"
    print "wsadmin.bat -lang jython -conntype NONE -wsadmin_classpath C:/WBI602/cloudscape/lib/db2j.jar" \
            " -f RelServiceUpgrade.py C:\\WBI602\\profiles\\WPSProfile C:\\backupWPSProfile"
    print "2. To manually upgrade a Cloudscape 5.1 instance:"
    print "wsadmin.bat -lang jython -conntype NONE -wsadmin_classpath C:/WBI602/cloudscape/lib/db2j.jar" \
            " -f RelServiceUpgrade.py -manuallyUpgrade C:\\WBI602\\profiles\\WPSProfile\\databases\\WPRCSDB"

#-----------------------------------------------------------------
# setup_Globals()
#
#-----------------------------------------------------------------
def setup_Globals():
   global AdminConfig
   
   global RELSERVICE_DEFAULT_DATASOURCE
   global RELATIONSHIP_INFRASTRUCTUREVERSION
   global ICS_VERSION
   global RELATIONSHIP_DATASOURCE
   global RELATIONSHIP_NAME
   global JDBCPROVIDER
   global JDBCPROVIDER_IMPLCLASSNAME
   global CLOUDSCAPE51_PACKAGE_PREFIX
   global DATASOURCE_JNDINAME
   global DATASOURCE_PROPERTYSET
   global DATASOURCE_RESOURCEPROPERTIES
   global DATASOURCE_J2EERESROUCEPROPERTY
   global DATASOURCE_RESOURCE_PROPERTY_NAME
   global PROPERTY_DATABASE_NAME
   global DATASOURCE_RESOURCE_PROPERTY_VALUE
   global ENVIRONMENT_VARIABLE_PREFIX
   global ENVIRONMENT_VARIABLE_POSTFIX
   global EMPTY_STRING
   
   RELSERVICE_DEFAULT_DATASOURCE = "defaultDataSource"
   RELATIONSHIP_INFRASTRUCTUREVERSION = "infrastructureVersion"
   ICS_VERSION = "4.3"
   RELATIONSHIP_DATASOURCE = "datasource"
   RELATIONSHIP_NAME = "relName"
   JDBCPROVIDER = "provider"
   JDBCPROVIDER_IMPLCLASSNAME = "implementationClassName"
   CLOUDSCAPE51_PACKAGE_PREFIX = "com.ibm.db2j"
   DATASOURCE_JNDINAME = "jndiName"
   DATASOURCE_PROPERTYSET = "propertySet"
   DATASOURCE_RESOURCEPROPERTIES = "resourceProperties"
   DATASOURCE_J2EERESROUCEPROPERTY = "J2EEResourceProperty"
   DATASOURCE_RESOURCE_PROPERTY_NAME = "name"
   PROPERTY_DATABASE_NAME = "databaseName"
   DATASOURCE_RESOURCE_PROPERTY_VALUE = "value"
   ENVIRONMENT_VARIABLE_PREFIX = "${"
   ENVIRONMENT_VARIABLE_POSTFIX = "}"
   EMPTY_STRING = ''


#-----------------------------------------------------------------
# Logs messages
#-----------------------------------------------------------------
def trace(message):
    print message


#-----------------------------------------------------------------
# Get JNDI names of data sources used by relationshipService
#-----------------------------------------------------------------
def getJNDINames():
   #-----------------------------------------------------------------
   # Get the RelationshipService WCCM model
   #-----------------------------------------------------------------
   trace ("To get JNDI names used in WPS60x: ")
   # Because RelationshipService a resource at cell-level, we don't specify here
   cfgObjs = AdminConfig.list("RelationshipService").split(lineSeparator)
   relService = cfgObjs[0]
   
   jndiNames = []
   defaultJNDI = AdminConfig.showAttribute(relService, 'defaultDataSource')
   trace ("Get JNDI name of the default data source: " + defaultJNDI + ".")
   jndiNames.append(defaultJNDI)
   
   # Get relationships in RelationshipService WCCM model
   relnList = AdminConfig.list("Relationship").split(lineSeparator)
   if len(relnList) == 0 :
       return jndiNames
   
   for reln in relnList :
       version = AdminConfig.showAttribute(reln, RELATIONSHIP_INFRASTRUCTUREVERSION)
       relationshipName = AdminConfig.showAttribute(reln, RELATIONSHIP_NAME)
       if version != '6.0' :
           # skip relationship that is not 6.0 version
           trace ("Skip relationship '"+ relationshipName + "' that is not of 6.0 version.")
       else :
           # get the data source used for relationship of DIAMOND_VERSION (6.0)
           name = AdminConfig.showAttribute(reln, RELATIONSHIP_DATASOURCE)
           if name not in jndiNames :
               trace ("Get JNDI name '"+ name + "' that is used for relationship '" + relationshipName + "'.")
               jndiNames.append(name)
   
   return jndiNames
   


#-----------------------------------------------------------------
# Get database URL
#-----------------------------------------------------------------
def getDatabaseURLInWPS60x(jndiName):
    trace ("To get database URL in WPS60x for JNDI '" + jndiName + "':")
    # read database url from DataSource
    dsList = AdminConfig.list("DataSource").split(lineSeparator)
    for ds in dsList :
        if ds != EMPTY_STRING:
            # 0> get the data source with the specified jndi name
            tmpJndiName = AdminConfig.showAttribute(ds, DATASOURCE_JNDINAME)
            if tmpJndiName != jndiName :
                continue
            # else, we get the specified data source
            # 1> tell whether this is a cloudscape data source
            jdbcProvider = AdminConfig.showAttribute(ds, JDBCPROVIDER)
            implClassName = AdminConfig.showAttribute(jdbcProvider, JDBCPROVIDER_IMPLCLASSNAME)
            trace ("The implementation class for data source '" + jndiName + "' is '" + implClassName + "'.")
            if not implClassName.startswith(CLOUDSCAPE51_PACKAGE_PREFIX) :
                trace ("Ignore data source '"+ jndiName +"' that is not for Cloudscape 5.1.")
                return EMPTY_STRING 
            else:
                trace ("Find a data source '" + jndiName + "' that is for Cloudscape 5.1.")
                # 2> get database URL from data source resource properties
                resrcPropList = AdminConfig.list(DATASOURCE_J2EERESROUCEPROPERTY, ds).split(lineSeparator)
                for resrcProp in resrcPropList :
                    propertyName = AdminConfig.showAttribute(resrcProp, DATASOURCE_RESOURCE_PROPERTY_NAME)
                    if propertyName == PROPERTY_DATABASE_NAME :
                        rawURL = AdminConfig.showAttribute(resrcProp, DATASOURCE_RESOURCE_PROPERTY_VALUE)
                        if rawURL.find(ENVIRONMENT_VARIABLE_PREFIX) != -1 :
                            parsedURL = parseEnvironmentVariable(rawURL)
                        else:
                            parsedURL = rawURL
                        trace ("Get Cloudscape 5.1 database URL '" + parsedURL + "' in WPS60x_HOME")
                        return parsedURL
              
    trace ("Data source '" + jndiName + "' is not found!")
    trace ("Please ensure this is a valid data source")
    return EMPTY_STRING        

#-----------------------------------------------------------------
# Parses environment variable(s) in the original database URL
#-----------------------------------------------------------------
def parseEnvironmentVariable(rawURL):
    # get environment variable
    numPrefix = rawURL.count(ENVIRONMENT_VARIABLE_PREFIX)
    numPostfix = rawURL.count(ENVIRONMENT_VARIABLE_POSTFIX)
    if numPrefix != numPostfix :
        trace (ENVIRONMENT_VARIABLE_PREFIX + " and " + ENVIRONMENT_VARIABLE_POSTFIX \
            + " don't match in this database URL '" + rawURL + "'")
        return
    else:
        # this is a valid database URL.
        iStart = rawURL.index(ENVIRONMENT_VARIABLE_PREFIX)
        iEnd = rawURL.index(ENVIRONMENT_VARIABLE_POSTFIX)
        while iStart != -1 :
            envVarName = rawURL[(iStart + 2) : iEnd]
            trace ("Environment variable name '" + envVarName + "'")
            envVarValue = getValue(envVarName)
            if envVarValue != '':
                if iStart > 0:
                    rawURL = rawURL[0 : iStart] + envVarValue + rawURL[(iEnd + 1) :]
                else:
                    rawURL = envVarValue + rawURL[(iEnd + 1) :]
            else:
                trace ("Environment variable '" + envVarName + "' is not found!")

            if rawURL.find(ENVIRONMENT_VARIABLE_PREFIX) != -1:
                iStart = rawURL.index(ENVIRONMENT_VARIABLE_PREFIX)
                iEnd = rawURL.index(ENVIRONMENT_VARIABLE_POSTFIX)
            else :
                iStart = -1
                iEnd = -1
        trace ("The parsed database URL is " + rawURL)
        return rawURL

#-----------------------------------------------------------------
# Returns the value of the environment variable
#-----------------------------------------------------------------
def getValue (name):
    variableLists = AdminConfig.list("VariableMap").split(lineSeparator)
    for eachList in variableLists :
        entries = AdminConfig.list("VariableSubstitutionEntry", eachList).split(lineSeparator)
        for entry in entries:
            if AdminConfig.showAttribute(entry, "symbolicName") == name:
                value = AdminConfig.showAttribute(entry, "value")
                trace ("The value of environment variable '" + name + "' is '" + value + "'")
                return value
    trace ("Environment variable '" + name + "' is not found!")
    return EMPTY_STRING


#-----------------------------------------------------------------
# Generates the corresponding database url in the backup directory
#-----------------------------------------------------------------
def getDatabaseURLInBackupDir(profileHome, backupDir, origURL):
    trace ("To get database URL in backup directory, profile home is '" + profileHome \
            + "', backup dir is '" + backupDir + "' and the original URL is '" + origURL +"'")
    # generate the corresponding database URL in the backup directory
    # for example:
    #    profileHome = 'c:\\WBI602\\profiles\\WPSProfile'
    #    backupDir = 'c:\\backup'
    #    origURL = 'c:\\WBI602\\profiles\\WPSProfile\\databases\\WPRCSDB'
    # then 'c:\\backup\\profiles\\WPSProfile\\databases\\WPRCSDB' is the target databaseURL
    
    # normalize file pathes
    profileHome = normalizePath(profileHome)
    backupDir = normalizePath(backupDir)
    origURL = normalizePath(origURL)
    wasHome = getWASHome()
    
    rc = 0 # return code
    profileName = getProfileName(profileHome)
    
    if (profileName == EMPTY_STRING):
      # failed to find a profile with path 'profileHome'
      rc = -1
    elif (origURL.find(profileHome) != -1):
      # the original database resides in profile home.
      index = origURL.find(profileHome) + len(profileHome) + 1
      databaseURL = backupDir + File.separator + "profiles" \
                            + File.separator + profileName \
                            + File.separator + origURL[index:]
      # validate the generated database URL
      if not existsPath(databaseURL):
          trace ("Directory does not exist: " + databaseURL);
          rc = -2
    elif (origURL.find(wasHome) != -1):
      # the original database resides in not in PROFILE_HOME but WAS_HOME
      # it will be copied to BACKUP_DIR/websphere_backup/...
      index = origURL.find(wasHome) + len(wasHome) + 1
      databaseURL = backupDir + File.separator + "websphere_backup" \
                            + File.separator + origURL[index:]
      # validate the generated database URL
      if not existsPath(databaseURL):
          trace ("Directory does not exist: " + databaseURL);
          rc = -2
    # failed to extract profile name. Please upgrade the Cloudscape5.1 instance manually
    else:
        rc = -3

    
    if rc == 0:
        trace ("The corresponding Cloudscape 5.1 instance in the backup directory is: '" + databaseURL + "'")
        return normalizePath(databaseURL)
    else:
        print "The return code is '" + rc.toString() + "'"
        print "Not able to generate the database URL in the backup directory for '" + origURL + "'"
        print "Please upgrade it manually like the following:"
        print "    wsadmin -lang jython -conntype NONE -wsadmin_classpath <WPS60x_HOME>/cloudscape/lib/db2.jar "\
              " -f RelServiceUpgrade.py -manuallyUpgrade <databaseURL>"

        return EMPTY_STRING
        


#-----------------------------------------------------------------
# Drops method aliases generated by relationshipService
#-----------------------------------------------------------------
def dropRelatedMethodAliases(databaseURL):
    conn = '' # Connection
    stmt = '' # Statement
    rs = ''   # ResultSet
    rc = 0    # return code
    try:
        # new a Cloudscape data source
        ds = DB2jDataSource()
        # set database name
        ds.setDatabaseName(databaseURL)
        
        try:
            # connect to the database
            conn = ds.getConnection()
            # get names of method aliases generated by relationshipService
            stmt = conn.createStatement()
            sql = "SELECT ALIAS FROM SYS.SYSALIASES WHERE JAVACLASSNAME LIKE 'com.ibm.wbiserver.relationshipservice.dblayer.sp.%' AND ALIASTYPE='M' AND NAMESPACE='M' AND SYSTEMALIAS=false"
            rs = stmt.executeQuery(sql)
    
            # construct SQL statements to drop method aliases generated by relationshipService
            sqlList = []
            while rs.next():
                methodAliasName = rs.getString(1)
                sqlList.append("DROP METHOD ALIAS " + methodAliasName)
        
            if len(sqlList) == 0:
                trace ("Method alias generated by RelationshipService component is not found. No action required for this database.")
                return
    
            for eachsql in sqlList :
                trace (eachsql)
                # drop method alias
                stmt.execute(eachsql)
        finally:
            if rs != '':
                rs.close()
            if stmt != '':
                stmt.close()
            if conn != '':
                conn.close()
    except:
        _type_, _value_, _tbck_ = sys.exc_info()
        error = str(_value_)
        rc = 1

    if (rc == 0):
        trace ("Done calling dropRelatedMethodAliases('" + databaseURL + "').")
    else:
        trace ("Error calling dropRelatedMethodAliases('" + databaseURL + "'): " + `error` + ".")
        trace (str(_type_))
        trace (str(_value_))
        trace (str(_tbck_))
   

#-----------------------------------------------------------------
# Path utilities
#-----------------------------------------------------------------
def existsPath(pathName):
 
  newPath=File(pathName)
  return newPath.exists()
 

def normalizePath(pathName):
  newPath=File(pathName)
  return newPath.getAbsolutePath()

#--------------------------------------------------------------------------
# get the directory where WAS is installed
#--------------------------------------------------------------------------
def getWASHome():
	return getValue('WAS_INSTALL_ROOT')


#--------------------------------------------------------------------------
# get profile name via its profileHome.
#--------------------------------------------------------------------------
def getProfileName(profileHome):
  dbFactory = DocumentBuilderFactory.newInstance()
  dBuilder = dbFactory.newDocumentBuilder()
  propertyFilePath = getWASHome() + File.separator + 'properties' + File.separator + 'profileRegistry.xml'
  profileRegistry = File(propertyFilePath)
  doc = dBuilder.parse(profileRegistry)
  profiles = doc.getElementsByTagName('profile')
  index = 0
  while(index < profiles.length):
    profile = profiles.item(index)
    path = profile.getAttribute('path')
    if(profileHome == normalizePath(path)):
      return profile.getAttribute('name')
    index = index + 1
  trace("Profile is not found with path '" + profileHome + "'.")
  return EMPTY_STRING
    

#-----------------------------------------------------------------
# Main
#-----------------------------------------------------------------
# 0> analysis the input arguments and parameters
length = len(sys.argv);
if (length == 1) and (sys.argv[0].upper() == "-help".upper()):
    print usage()

elif length == 2:
    if sys.argv[0].upper() == "-manuallyUpgrade".upper():
        # it is used to manually upgrade a Cloudscape5.1 instance
        # check that <databaseURL> is valid
        dbpath = sys.argv[1]
        if not existsPath(dbpath):
            print "Invalid database URL: '" + dbpath + "'!"
        else:
            # this is a valid file path
            print "To drop method aliases in Cloudscape5.1 instance '" + dbpath + "'"
            dropRelatedMethodAliases ( normalizePath(dbpath) )

    else:
        # it is used in profile pre-upgrade
        # check that <WPS60x_PROFILE_HOME> is valid
        # check that <backup_directory> is valid
        WPS60x_PROFILE_HOME = sys.argv[0]
        backupDir = sys.argv[1]        
        if not existsPath(WPS60x_PROFILE_HOME):
            print "Invalid WPS60x_PROFILE_HOME '" + WPS60x_PROFILE_HOME + "'!"
        if not existsPath(backupDir):
            print "Invalid backup directory '" + backupDir + "'!"
        else:
            # valid arguments
            setup_Globals()
            # 1> get JNDI names of data sources used by RelationshipService
            jndiNames = getJNDINames()
    
            if len(jndiNames) == 0 :
                # this should not happen
                trace ("Data source used by RelationshipService is not found!")
            else:
                for jndiName in jndiNames :
                    if(jndiName == EMPTY_STRING):
                        continue
                    # 2> get database URL that recorded in the specified data source in WPS60x_PROFILE_HOME
                    trace ("\nTo process data source with JNDI name '" + jndiName + "':")
                    rawURL = getDatabaseURLInWPS60x(jndiName)
                    if rawURL != EMPTY_STRING:
                        # 3> get the corresponding database URL in the backup direcotry
                        databaseURL = getDatabaseURLInBackupDir(WPS60x_PROFILE_HOME, backupDir, rawURL)
                        if databaseURL != EMPTY_STRING:
                            trace ("To drop method aliases in Cloudscape5.1 instance '" + databaseURL + "'")
                            dropRelatedMethodAliases(databaseURL)

else:
    print usage()
